"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("../../utilities/graphql");
const helpers_1 = require("./helpers");
const t = require("@babel/types");
class TypescriptGenerator {
    constructor(compilerOptions) {
        this.options = compilerOptions;
        this.typeFromGraphQLType = helpers_1.createTypeFromGraphQLTypeFunction(compilerOptions);
    }
    scalarDeclaration(scalarType) {
        const { name, description } = scalarType;
        const scalarDeclaration = t.exportNamedDeclaration(t.TSTypeAliasDeclaration(t.identifier(name), undefined, t.TSAnyKeyword()), []);
        if (description) {
            const leadingComment = helpers_1.convertStringToDocstringComment(description);
            if (leadingComment) {
                scalarDeclaration.leadingComments = [leadingComment];
            }
        }
        return scalarDeclaration;
    }
    enumerationDeclaration(type) {
        const { name, description } = type;
        const enumMembers = graphql_1.sortEnumValues(type.getValues()).map(({ value }) => {
            return t.TSEnumMember(t.identifier(value), t.stringLiteral(value));
        });
        const typeAlias = t.exportNamedDeclaration(t.TSEnumDeclaration(t.identifier(name), enumMembers), []);
        if (description) {
            const leadingComment = helpers_1.convertStringToDocstringComment(description);
            if (leadingComment) {
                typeAlias.leadingComments = [leadingComment];
            }
        }
        return typeAlias;
    }
    inputObjectDeclaration(inputObjectType) {
        const { name, description } = inputObjectType;
        const fieldMap = inputObjectType.getFields();
        const fields = Object.keys(inputObjectType.getFields())
            .map((fieldName) => {
            const field = fieldMap[fieldName];
            return {
                name: fieldName,
                type: this.typeFromGraphQLType(field.type)
            };
        });
        const inputType = this.exportDeclaration(this.interface(name, fields, {
            keyInheritsNullability: true
        }));
        if (description) {
            const leadingComment = helpers_1.convertStringToDocstringComment(description);
            if (leadingComment) {
                inputType.leadingComments = [leadingComment];
            }
        }
        return inputType;
    }
    typesForProperties(fields, { keyInheritsNullability = false } = {}) {
        return fields.map(({ name, description, type }) => {
            const propertySignatureType = t.TSPropertySignature(t.identifier(name), t.TSTypeAnnotation(type));
            propertySignatureType.optional = keyInheritsNullability && this.isNullableType(type);
            if (description) {
                const leadingComment = helpers_1.convertStringToDocstringComment(description);
                if (leadingComment) {
                    propertySignatureType.leadingComments = [leadingComment];
                }
            }
            return propertySignatureType;
        });
    }
    interface(name, fields, { keyInheritsNullability = false } = {}) {
        return t.TSInterfaceDeclaration(t.identifier(name), undefined, undefined, t.TSInterfaceBody(this.typesForProperties(fields, {
            keyInheritsNullability
        })));
    }
    typeAliasGenericUnion(name, members) {
        return t.TSTypeAliasDeclaration(t.identifier(name), undefined, t.TSUnionType(members));
    }
    exportDeclaration(declaration) {
        return t.exportNamedDeclaration(declaration, []);
    }
    nameFromScopeStack(scope) {
        return scope.join('_');
    }
    makeNullableType(type) {
        return t.TSUnionType([
            type,
            t.TSNullKeyword()
        ]);
    }
    isNullableType(type) {
        return t.isTSUnionType(type) && type.types.some(type => t.isTSNullKeyword(type));
    }
}
exports.default = TypescriptGenerator;
//# sourceMappingURL=language.js.map